from django.core.exceptions import ObjectDoesNotExist

from isfilled.models import Fill
from isfilled.util import import_string

class Filled(object):
    name = None
    fields = []
    exclude = []

    def __init__(self, instance=None):
        self.instance = instance

    def default_fields(self):
        if self.instance.fill:
            res = self.form().fields.keys()
        elif self.instance.form:
            res = self.get_form_fields(self.instance.using_form())
        else:
            res = self.get_model_fields(self.instance.model_model())
        return res

    def using_fields(self, keys, exclude=[]):
        return list(set(keys) - set(exclude))

    def is_field_filled(self, instance, field):
        # not all form fields are model fields; ignore
        if not hasattr(instance, field):
            return True
        return getattr(instance, field)

    def verify(self, instance, fields):
        return all(self.is_field_filled(instance, k) for k in fields)

    def registered_model_fills(self):
        return Fill.objects.filter(model=self.instance._meta.label)

    def as_fills(self, fills):
        return [import_string(fill.fill) if fill.fill else None for fill in fills]

    def get_model_fields(self, instance):
        return [k.name for k in instance._meta.get_fields()]

    def get_form_fields(self, form):
        return form().fields.keys()

    def str_as_list(self, fld):
        return filter(None, (fld or '').split(','))

class FillsMixin(object):
    """ Fills superpowers for Models """

    def check_fills(self, fills=[]):
        dummy = Filled(instance=self)
        registered_model_fills = fills or dummy.registered_model_fills()
        registered_fills = dummy.as_fills(registered_model_fills)
        def _fields(fill, instance):
            return fill.using_fields(keys=instance.str_as_list('fields') or fill.default_fields(),
                               exclude=instance.str_as_list('exclude'))
        def _verify(fill, instance):
            fill_instance = fill(instance=instance) if fill else Filled(instance=instance)
            flds = _fields(fill_instance, instance)
            return fill_instance.verify(self, flds)

        ctxs = {fill: _verify(fill, ins) for fill, ins in 
                    zip(registered_fills, registered_model_fills)}
        return type('Result', (object,), {'state': all(ctxs.values()), 'contexts': ctxs})

    def is_filled(self, form=None, fill=None):
        dummy = Filled(instance=self)
        if form is None and fill is None:
            fields = dummy.get_model_fields(self)
        elif form:
            fields = dummy.get_form_fields(form)
        elif fill:
            try:
                db = Fill.objects.get(name=fill.name,
                                      model=fill.form.Meta.model._meta.label)
            except ObjectDoesNotExist:
                db = None
            fields = dummy.using_fields(
                    (dummy.str_as_list(db.fields) if db else fill.fields) or dummy.get_form_fields(fill.form),
                    (dummy.str_as_list(db.exclude) if db else fill.exclude) or fill.exclude,)
        else:
            raise Exception("The impossible happened")

        return dummy.verify(self, fields)

        




from django.conf import settings
from django.contrib import admin

from isfilled.models import Fill

class FillAdmin(admin.ModelAdmin):
    list_display = ['name', 'fill', 'model', 'form',]

admin.site.register(Fill, FillAdmin)



# -*- coding: utf-8 -*-
from django.apps import apps
from django.core.cache import cache
from django.conf import settings
from django.db import models
from django.db.models import Count

from isfilled.util import import_string

class Fill(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    name = models.CharField(max_length=255, db_index=True)
    fill = models.CharField(max_length=255, null=True, blank=True, db_index=True)
    model = models.CharField(max_length=255, null=True, blank=True, db_index=True)
    form = models.CharField(max_length=255, null=True, blank=True)
    fields = models.CharField(max_length=255, null=True, blank=True)
    exclude = models.CharField(max_length=255, null=True, blank=True)

    def str_as_list(self, fld):
        return filter(None, (getattr(self, fld) or '').split(','))

    def fill_model(self, fill=None):
        fill = fill or import_string(self.fill)
        return fill.form.Meta.model

    def form_model(self):
        return self.using_form().Meta.model

    def using_form(self):
        return import_string(self.form)

    def model_model(self):
        return apps.get_model(self.model)

    def save(self, *args, **kwargs):
        if self.fill:
            fi = import_string(self.fill)
            self.name = fi.name
            if not self.model:
                self.model = self.fill_model(fi)._meta.label
        if self.form and not self.model:
            self.model = self.form_model()._meta.label
        super(Fill, self).save(*args, **kwargs)

    def __unicode__(self):
        return self.name

    class Meta:
        unique_together = (('name','model'),)



# -*- coding: utf-8 -*-
from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.db import models
from django.test import TestCase, TransactionTestCase, override_settings
from django.test.client import Client
from django.contrib.auth.models import Group, User
from django import forms

from isfilled.models import Fill
from isfilled.fills import FillsMixin, Filled

import os, logging, datetime
logging.basicConfig(level=logging.DEBUG)

class BaseSuite(TransactionTestCase):
    pass

class Employee(models.Model, FillsMixin):
    name = models.CharField(max_length=255)
    expert_in = models.CharField(max_length=255, blank=True, null=True)
    starting_date = models.DateField(blank=True, null=True)
    editor_of_choice = models.CharField(max_length=255, blank=True, null=True)

class EmployeeFillsForm(forms.ModelForm):
    class Meta:
        model = Employee
        exclude = []

class EmployeeFillsExcludeForm(forms.ModelForm):
    class Meta:
        model = Employee
        exclude = ['editor_of_choice','starting_date',]

class EditorForm(forms.ModelForm):
    class Meta:
        model = Employee
        fields = ['editor_of_choice']

class EmployeeFills(Filled):
    name = 'employee-fills'
    form = EmployeeFillsForm

class EmployeeExpertFills(Filled):
    form = EmployeeFillsForm
    fields = ['expert_in',]

class EmployeeExpertExcludeFills(Filled):
    form = EmployeeFillsForm
    exclude = ['expert_in',]

class FillsTest(BaseSuite):

    def test_fills(self):
        e = Employee.objects.create(name="Jared")

        f = Fill.objects.create(
                name="employee-hr",
                fill="test.test_fills.EmployeeFills",)

        self.assertEqual(e.check_fills().state, False)

        e.expert_in = "iOS"
        e.starting_date = datetime.date.today()
        e.editor_of_choice = 'vim'
        e.save()

        self.assertEqual(e.check_fills().state, True)

    def test_model_fills(self):
        e = Employee.objects.create(name="Jared")

        f = Fill.objects.create(
                name="employee",
                model="test.Employee",)

        self.assertEqual(e.check_fills().state, False)

        e.expert_in = "Android C++ C#"
        e.starting_date = datetime.date.today()
        e.editor_of_choice = 'vim'
        e.save()

        self.assertEqual(e.check_fills().state, True)

    def test_form_fills(self):
        e = Employee.objects.create(name="Jared")

        f = Fill.objects.create(
                name="employee",
                form="test.test_fills.EmployeeFillsForm",)

        self.assertEqual(e.check_fills().state, False)

        e.expert_in = "Android C++ C#"
        e.save()

        self.assertEqual(e.check_fills().state, False)

        e.starting_date = datetime.date.today()
        e.save()

        self.assertEqual(e.check_fills().state, False)

        e.editor_of_choice = 'vim'
        e.save()

        self.assertEqual(e.check_fills().state, True)

    def test_selected_form_fills(self):
        e = Employee.objects.create(name="Jared")

        f = Fill.objects.create(
                name="employee",
                form="test.test_fills.EditorForm",)

        self.assertEqual(e.check_fills().state, False)

        e.editor_of_choice = "emacs"

        self.assertEqual(e.check_fills().state, True)

    def test_selected_form_fills_noop(self):
        e = Employee.objects.create(name="Jared")

        f = Fill.objects.create(
                name="employee",
                form="test.test_fills.EditorForm",
                exclude="editor_of_choice")

        self.assertEqual(e.check_fills().state, True)

    def test_model_is_filled(self):
        e = Employee.objects.create(name="Jared")
        self.assertFalse(e.is_filled())
        e.expert_in = "Android C++ C#"
        e.starting_date = datetime.date.today()
        e.editor_of_choice = 'vim'
        self.assertTrue(e.is_filled())

    def test_model_is_filled_given_a_form(self):
        e = Employee.objects.create(name="Jared")
        self.assertFalse(e.is_filled(form=EmployeeFillsForm))
        e.expert_in = "Android C++ C#"
        e.starting_date = datetime.date.today()
        e.editor_of_choice = 'vim'
        self.assertTrue(e.is_filled(form=EmployeeFillsForm))

    def test_model_is_filled_given_a_form_exclude(self):
        e = Employee.objects.create(name="Jared")
        self.assertFalse(e.is_filled(form=EmployeeFillsExcludeForm))
        e.expert_in = "Android C++ C#"
        self.assertTrue(e.is_filled(form=EmployeeFillsExcludeForm))

    def test_model_is_filled_given_a_fill(self):
        e = Employee.objects.create(name="Jared")

        f = Fill.objects.create(fill="test.test_fills.EmployeeFills", exclude="editor_of_choice")

        self.assertFalse(e.is_filled(fill=EmployeeFills))

        e.expert_in = "Android C++ C#"
        e.starting_date = datetime.date.today()
        self.assertTrue(e.is_filled(fill=EmployeeFills))

    def test_model_is_filled_given_a_fill_expert_fields(self):
        e = Employee.objects.create(name="Jared")
        self.assertFalse(e.is_filled(fill=EmployeeExpertFills))
        e.expert_in = "food"
        self.assertTrue(e.is_filled(fill=EmployeeExpertFills))

    def test_model_is_filled_given_a_fill_expert_exclude(self):
        e = Employee.objects.create(name="Jared")
        self.assertFalse(e.is_filled(fill=EmployeeExpertExcludeFills))
        e.expert_in = "food"
        self.assertFalse(e.is_filled(fill=EmployeeExpertExcludeFills))
        e.editor_of_choice = 'vim'
        e.expert_in = None
        e.starting_date = datetime.date.today()
        self.assertTrue(e.is_filled(fill=EmployeeExpertExcludeFills))
